package be.iminds.ilabt.jfed.restapi.jfedclient;

import be.iminds.ilabt.jfed.*;
import be.iminds.ilabt.jfed.experiment.ExperimentControllerFactory;
import be.iminds.ilabt.jfed.experiment.ExperimentController;
import be.iminds.ilabt.jfed.experiment.Experiment;
import be.iminds.ilabt.jfed.experiment.ExperimentState;
import be.iminds.ilabt.jfed.fedmon.webapi.service.json.Server;
import be.iminds.ilabt.jfed.highlevel.HighLevelModule;
import be.iminds.ilabt.jfed.highlevel.SfaOnlyExperimentModule;
import be.iminds.ilabt.jfed.highlevel.controller.*;
import be.iminds.ilabt.jfed.highlevel.model.SfaModel;
import be.iminds.ilabt.jfed.highlevel.model.Slice;
import be.iminds.ilabt.jfed.highlevel.model.UserSpecHelper;
import be.iminds.ilabt.jfed.highlevel.tasks.AggregatesForSliceTask;
import be.iminds.ilabt.jfed.highlevel.tasks.HighLevelTaskFactory;
import be.iminds.ilabt.jfed.highlevel.tasks.ProjectMemberSshKeysTask;
import be.iminds.ilabt.jfed.highlevel.tasks.SliceManifestForAuthorityTask;
import be.iminds.ilabt.jfed.highlevel.util.AggregateManagerWrapperFactory;
import be.iminds.ilabt.jfed.log.ApiCallDetails;
import be.iminds.ilabt.jfed.log.Logger;
import be.iminds.ilabt.jfed.log.ResultListener;
import be.iminds.ilabt.jfed.lowlevel.*;
import be.iminds.ilabt.jfed.lowlevel.api.user_spec.UserSpec;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.StatusDetails;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.UserAndSliceApiWrapper;
import be.iminds.ilabt.jfed.lowlevel.authority.AuthorityFinder;
import be.iminds.ilabt.jfed.lowlevel.userloginmodel.InvalidLoginException;
import be.iminds.ilabt.jfed.restapi.debugdata.ClientApiCallDDBuilder;
import be.iminds.ilabt.jfed.restapi.debugdata.ClientApiCallDDCollector;
import be.iminds.ilabt.jfed.rspec.model.ModelRspecType;
import be.iminds.ilabt.jfed.rspec.model.StringRspec;
import be.iminds.ilabt.jfed.rspec.rspec_source.ManifestRspecSource;
import be.iminds.ilabt.jfed.rspec.rspec_source.RequestRspecSource;
import be.iminds.ilabt.jfed.util.GeniUrn;
import be.iminds.ilabt.jfed.util.IOUtils;
import be.iminds.ilabt.jfed.util.JFedTrustStore;
import com.google.common.util.concurrent.ListenableFuture;
import com.google.inject.Guice;
import com.google.inject.Injector;
import org.slf4j.LoggerFactory;

import javax.annotation.Nullable;
import javax.ws.rs.ServiceUnavailableException;
import javax.ws.rs.WebApplicationException;
import javax.ws.rs.core.Response;
import java.io.*;
import java.sql.Timestamp;
import java.time.Instant;
import java.time.ZonedDateTime;
import java.util.*;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.stream.Collectors;

import static javax.xml.stream.XMLEventFactory.newInstance;

/**
 * JFedExperimenterApi: API offering jFed experimenter functionality.
 */
public class JFedExperimenterApiImpl implements JFedExperimenterApi {
    private static final org.slf4j.Logger LOG = LoggerFactory.getLogger(JFedExperimenterApiImpl.class);

    private final Logger logger;
    private final GeniUser user;
    private final TestbedInfoSource testbedInfoSource;
    private final AuthorityFinder authorityFinder;
    private final JFedTrustStore jFedTrustStore;
    private final JavaFXTaskThread javaFXTaskThread;
    private final TaskThread taskThread;
    private final SfaModel sfaModel;
    private final HighLevelTaskFactory hltf;
    private final Injector injector;

    public JFedExperimenterApiImpl(File userLoginPemKeyCertFile, char[] pass, List<AnyCredential> speaksForCredentials, @Nullable GeniUrn speakingForUrn)
            throws IOException, SimpleGeniUserWithSpeaksFor.InvalidSpeaksForException, InvalidLoginException {
        //////////////////////////////////////////// Setup user ////////////////////////////////////////////////////

        assert userLoginPemKeyCertFile != null;
        //allowed to be null in certain cases
//        assert speaksForCredentials != null;
//        assert !speaksForCredentials.isEmpty();

        String pemFileContent = IOUtils.fileToString(userLoginPemKeyCertFile);
        //this doesn;t work anymore, since DirectUserModule does not work anymore
//        SimpleGeniUserWithSpeaksFor inputUser = new SimpleGeniUserWithSpeaksFor(null, null, pemFileContent,
//                (pass == null || pass.length == 0) ? null : pass, userLoginPemKeyCertFile, userLoginPemKeyCertFile, speaksForCredentials);

        ////////////////////////////////////////// Get target test authority ///////////////////////////////////////

        injector = Guice.createInjector(
                new BasicTestbedInfoModule(),
                new JFedWebApiClientModule(),
                new KeyCertFileUserModule(
                        pemFileContent,
                        ((pass == null || pass.length == 0) ? (char[])null : pass),
                        userLoginPemKeyCertFile,
                        userLoginPemKeyCertFile,
                        speaksForCredentials),
                new PreLoginCoreModule(),
                new PostLoginCoreModule(),
                new CorePreferencesModule(),
                new HighLevelModule(),
                new SfaOnlyExperimentModule());
        testbedInfoSource =  injector.getInstance(TestbedInfoSource.class);
        authorityFinder =  injector.getInstance(AuthorityFinder.class);
        sfaModel = injector.getInstance(SfaModel.class);
        hltf = injector.getInstance(HighLevelTaskFactory.class);
        javaFXTaskThread = injector.getInstance(JavaFXTaskThread.class);
        taskThread = injector.getInstance(TaskThread.class);
        logger = injector.getInstance(Logger.class);
        jFedTrustStore = injector.getInstance(JFedTrustStore.class);
        user = injector.getInstance(GeniUser.class);

        if (user != null) {
            LOG.debug("You are logged in as user: " + user.getUserUrn());
            if (user.getUserAuthorityServer() != null) {
                LOG.debug("             on authority: " + user.getUserAuthorityServer().getDefaultComponentManagerUrn());
                if (user.getUserAuthorityServer().getServices() == null || user.getUserAuthorityServer().getServices().isEmpty()) {
                    LOG.error("ERROR: there are no URLs known for your user authority! (services=" + user.getUserAuthorityServer().getServices() + ")");
                    System.exit(1);
                    throw new RuntimeException("ERROR: there are no URLs known for your user authority! (services=" + user.getUserAuthorityServer().getServices() + ")");
                }
            } else {
                LOG.error("ERROR: the authority of your user was not found!");
                System.exit(1);
                throw new RuntimeException("");
            }
        } else {
            LOG.error("ERROR: null user!");
            throw new RuntimeException("ERROR: null user!");
        }

        UserAndSliceApiWrapper userAndSliceApiWrapper = injector.getInstance(UserAndSliceApiWrapper.class);
        AggregateManagerWrapperFactory aggregateManagerWrapperFactory = injector.getInstance(AggregateManagerWrapperFactory.class);

        //not needed anymore: this should be automatically done now
//        if (speaksForCredentials != null && !speaksForCredentials.isEmpty()) {
//            userAndSliceApiWrapper.setSpeaksFor(speaksForCredentials, speakingForUrn);
//            aggregateManagerWrapperFactory.setSpeaksFor(speaksForCredentials, speakingForUrn);
//        }
        if (speaksForCredentials != null && !speaksForCredentials.isEmpty()) {
            if (!userAndSliceApiWrapper.isSpeaksFor())
                throw new RuntimeException("Internal bug: speaksFor not enabled for userAndSliceApiWrapper");
            //not activated automatically for the factory, but for each AutomaticAMWrapper. So we can't check this.
//            if (!aggregateManagerWrapperFactory.isSpeaksFor())
//                throw new RuntimeException("Internal bug: speaksFor not enabled for aggregateManagerWrapperFactory");
        }

//        Server wall = authorityFinder.findByUrn("urn:publicid:IDN+wall2.ilabt.iminds.be+authority+cm", AuthorityFinder.Purpose.CREATE_SLIVER);
//
//        ///////////////////////////////////////////// Call any GetVersion /////////////////////////////////////////
//
//        JFedConnectionProvider conPool = injector.getInstance(JFedConnectionProvider.class);
//        JFedPreferences jFedPreferences = injector.getInstance(JFedPreferences.class);
//                   AggregateManagerWrapper amWrapper = injector.getInstance(AutomaticAggregateManagerWrapper.AutomaticAggregateManagerWrapperFactory.class).
//                           create(wall);
//                   UserAndSliceApiWrapper credWrapper = injector.getInstance(AutomaticUserAndSliceApiWrapper.AutomaticUserAndSliceApiWrapperFactory.class).
//                           create();
//
//                   amWrapper.getVersion();
    }

    @Override
    public Logger getLogger() {
        return logger;
    }

    @Override
    public GeniUser getUser() {
        return user;
    }

    @Override
    public TestbedInfoSource getTestbedInfoSource() {
        return testbedInfoSource;
    }

    @Override
    public AuthorityFinder getAuthorityFinder() {
        return authorityFinder;
    }

    @Override
    public JFedTrustStore getJFedTrustStore() {
        return jFedTrustStore;
    }


    private Experiment experiment;
    private ExperimentController experimentController;


    private final List<ClientApiCallDDCollector> clientApiCallDDCollectorList = new ArrayList<>();
    @Override
    public boolean removeClientApiCallDDCollector(ClientApiCallDDCollector clientApiCallDDCollector) {
        return clientApiCallDDCollectorList.remove(clientApiCallDDCollector);
    }
    @Override
    public void addClientApiCallDDCollector(ClientApiCallDDCollector clientApiCallDDCollector) {
        this.clientApiCallDDCollectorList.add(clientApiCallDDCollector);
    }

    @Override
    public void setClientApiCallDDCollector(ClientApiCallDDCollector clientApiCallDDCollector) {
        assert clientApiCallDDCollector != null;
        this.clientApiCallDDCollectorList.clear();
        this.clientApiCallDDCollectorList.add(clientApiCallDDCollector);
    }

    private ResultListener activeResultListener;
    private void initLogging() {
        LOG.debug("initLogging()");
        if (activeResultListener != null)
            return;

        activeResultListener =
                new ResultListener() {
                    @Override
                    public void onResult(ApiCallDetails apiCallDetails) {
                        if (clientApiCallDDCollectorList.isEmpty())
                            return;

                        ClientApiCallDDBuilder clientApiCallDD = new ClientApiCallDDBuilder(apiCallDetails);
                        for (ClientApiCallDDCollector clientApiCallDDCollector : clientApiCallDDCollectorList) {
                            clientApiCallDDCollector.addClientApiCall(clientApiCallDD);
                        }
                    }
                };
        logger.addResultListener(activeResultListener);
    }

    @Override
    public List<GeniUrn> getProjectsForUser(GeniUrn user) {
        assert user != null;
        UserAndSliceApiWrapper userAndSliceApiWrapper = injector.getInstance(UserAndSliceApiWrapper.class);

        List<String> subAuthNames;
        try {
            List check = userAndSliceApiWrapper.getCachedUserCredentialsForAM();
            if (check == null || check.isEmpty())
                userAndSliceApiWrapper.getUserCredentials(logger, user);

            subAuthNames = userAndSliceApiWrapper.getSubAuthorityNames(logger, user);
        } catch (JFedException e) {
            LOG.error("Failed to fetch projects", e);
            throw new WebApplicationException("Failed to fetch projects", Response.Status.INTERNAL_SERVER_ERROR);
        }
        List<GeniUrn> res = new ArrayList<>();
        for (String subAuthName : subAuthNames) {
            GeniUrn projectUrn = GeniUrn.createGeniUrnFromDecodedParts(user.getTopLevelAuthority_withoutSubAuth(), "project", subAuthName);
            res.add(projectUrn);
        }

        return res;
    }

    @Override
    public List<GeniUrn> getSlicesForUser(GeniUrn user) {
        UserAndSliceApiWrapper userAndSliceApiWrapper = injector.getInstance(UserAndSliceApiWrapper.class);
        List<GeniUrn> res;
        try {
            List check = userAndSliceApiWrapper.getCachedUserCredentialsForAM();
            if (check == null || check.isEmpty())
                userAndSliceApiWrapper.getUserCredentials(logger, user);

            res = userAndSliceApiWrapper.getSlicesForUser(logger, user);
        } catch (JFedException e) {
            LOG.error("Failed to fetch slices", e);
            throw new WebApplicationException("Failed to fetch slices", Response.Status.INTERNAL_SERVER_ERROR);
        }

        return res;
    }

    @Override
    public void createSliceAndSlivers(String sliceName, String project, String rspec, Instant expire) {
        RequestRspecSource newRequestRspecSource = new RequestRspecSource(rspec, ModelRspecType.BASIC);
        Instant requestedStartTime = null /*only for reservations*/;
        Instant requestedEndTime;
        if (expire == null) {
            requestedEndTime = ZonedDateTime.now()
                    .plusDays(0)
                    .plusHours(2)
                    .plusMinutes(0)
                    .plusSeconds(0).toInstant(); //=expiration time
        } else {
            requestedEndTime = expire;
        }
        boolean includeProjectMemberSshKeys = false;
        boolean requestSla = false;

        assert sliceName != null;
        assert newRequestRspecSource != null;
        assert requestedEndTime != null;

        initLogging();
        List<UserSpec> userSpecs = null; //note: if this stays null and is passed to "new Experiment",
        //      StartExperimentJob will add the SSH keys of the correct user,
        //      and those in the RSpec automatically.
        if (includeProjectMemberSshKeys && project != null) {
            ProjectMemberSshKeysTask task = hltf.getProjectMembersSSHKeys(project);

            if (task.newExecutionNeeded()) {
                LOG.info("Requesting project member info before continuing");

                TaskExecution<ProjectMemberSshKeysTask> taskExecution = taskThread.addTask(task);

                try {
                    while (!taskExecution.isCompleted()) {
                        if (task.getTotalMembersCount() != -1) {
                            LOG.info(String.format("Requesting project members info (%d/%d)",
                                    task.getMemberInfoRequested(), task.getTotalMembersCount()));
                        }

                        Thread.sleep(100);
                    }
                } catch (InterruptedException e) {
                    LOG.warn("Interrupted while requesting project member info");
                    Thread.currentThread().interrupt();
                }
            }

            userSpecs = UserSpecHelper.getProjectMemberUserSpecs(project, sfaModel);
        }
        experiment = new Experiment(sliceName, project, newRequestRspecSource,
                requestedStartTime, requestedEndTime,
                userSpecs, requestSla);

        ExperimentControllerFactory experimentControllerFactory = injector.getInstance(ExperimentControllerFactory.class);
        experimentController = experimentControllerFactory.createExperimentController(experiment);

        experimentController.start();

        long maxWaitTimeMs = 9000l;
        long startTime = System.currentTimeMillis();
        while (experiment.getExperimentState() == ExperimentState.PENDING || experiment.getExperimentState() == ExperimentState.RESTORING) {
            try {
                Thread.sleep(250);
                long curTime = System.currentTimeMillis();
                long diff = curTime - startTime;
                if (diff >= maxWaitTimeMs) {
                    LOG.warn("Waiting for slice creation took longer than "+(maxWaitTimeMs/1000.0)+" seconds: aborting wait and returning partial result to user.");
                    break;
                }
            } catch (InterruptedException e) {
                LOG.warn("Sleep interrupted. Will return partial result to user.");
                return;
            }
        }

        if (experiment.getExperimentState() == ExperimentState.FAILED) {
            throw new WebApplicationException("Failed to create slice", Response.Status.INTERNAL_SERVER_ERROR);
        }

        assert experimentController != null : "experimentController == null";
        assert experiment != null : "experiment == null";

        //we might have aborted the wait, so don't assume this
//        assert experiment.getSlice() != null : "experiment.getSlice() == null";
    }

    private Date previousRefreshTime = null;
    @Override
    public void refreshStatusIfNeeded() {
        if (previousRefreshTime != null) {
            //5 seconds in between refreshes at least
            Date earliestNewRefreshDate = new Date(previousRefreshTime.getTime() + 5 * 1000l);
            if (earliestNewRefreshDate.after(new Date())) {
                LOG.debug("refreshStatusIfNeeded() called to soon. Will reply with SERVICE UNAVAILABLE with Retry-After: 5");
                throw new ServiceUnavailableException(
                        "Cannot refresh status so soon after previous refresh. Wait a few seconds before retrying.",
                        5l/*retryAfter in seconds*/);
            }
        }

        switch (experiment.getExperimentState()) {
            case PENDING:      //fall-through
            case RESTORING:    //fall-through
            case PROVISIONING: //fall-through
            case TESTING:      //fall-through
            case DELETING:
            {
                LOG.debug("refreshStatusIfNeeded() when experiment.getExperimentState()="+
                        experiment.getExperimentState()+" -> will not refresh now, that will be done automatically.");
                return;
            }
            //do refresh in all other cases
            default: break;
        }

        LOG.debug("refreshStatusIfNeeded() will request refresh of status and wait for it up to 5 seconds");

        //this will prevent refresh status during this call.
        previousRefreshTime = new Date();

        ListenableFuture<Void> future = experimentController.requestUpdate();

        long maxWaitTimeMs = 5000l;
        long startTime = System.currentTimeMillis();
        while (!(future.isDone() || future.isCancelled())) {
            try {
                Thread.sleep(100);
                long curTime = System.currentTimeMillis();
                long diff = curTime - startTime;
                if (diff >= maxWaitTimeMs) {
                    LOG.warn("Waiting for status update took longer than 5 seconds: aborting wait and returning current status to user.");
                    break;
                }
            } catch (InterruptedException e) {
                LOG.warn("Sleep interrupted. Will return current status to user.");
                return;
            }
        }

        //this will prevent refresh status too soon after this one finished
        previousRefreshTime = new Date();
    }

    @Override
    public void recoverSlice(GeniUrn sliceUrn) {
        initLogging();

        Slice recoverSlice = sfaModel.logExistSlice(sliceUrn);

        AggregatesForSliceTask getAggregatesForSliceTask =
                hltf.getAggregatesForSlice(recoverSlice);

        final TaskExecution taskExecution = hltf.submitTask(getAggregatesForSliceTask);

        while (!taskExecution.isCompleted()) {
            try {
                Thread.sleep(100);
            } catch (InterruptedException e) {
                LOG.warn("Sleep interrupted. Will abort slice recovery");
                return;
            }
        }

        if (taskExecution.getState() != TaskExecution.TaskState.SUCCESS) {
            throw new RuntimeException("Error recovering slice");
        }

        boolean usableSliverInfo = true;
        List<GeniUrn> aggregateUrns = getAggregatesForSliceTask.getAggregates();
        Set<Server> involvedAuthorities;
        if (aggregateUrns != null) {
            involvedAuthorities =
                    aggregateUrns.stream()
                            .map(urn -> testbedInfoSource.getByUrnExact(urn))
                            .filter(o -> o != null)
                            .collect(Collectors.toSet());
            List<GeniUrn> unknownAuthorities = aggregateUrns.stream()
                    .filter(urn -> testbedInfoSource.getByUrnExact(urn) == null)
                    .collect(Collectors.toList());

            LOG.info("Recover: got SLIVER_INFO: involvedAuthorities=" + involvedAuthorities);

            if (involvedAuthorities.isEmpty())
                usableSliverInfo = false;

            if (!unknownAuthorities.isEmpty()) {
                LOG.warn("Found unknown authorities while recovering {}: {}", recoverSlice.getName(), unknownAuthorities);
                throw new RuntimeException("Found unknown authorities while recovering "+recoverSlice.getName()+": "+unknownAuthorities);
            }
        } else {
            usableSliverInfo = false;
            involvedAuthorities = null;
        }

        if (!usableSliverInfo) {
            LOG.warn("Could not recover slice {} using SLIVER_INFO", recoverSlice.getName());
            //will not see this as error. result will be Experiment in NON_EXISTING state
//            throw new RuntimeException("Could not recover slice "+recoverSlice.getName()+" using SLIVER_INFO");
        } else {
            final Set<Server> initialInvolvedAuthorities = new HashSet<>(involvedAuthorities);
            final AtomicInteger finishedTaskExecutionsCount = new AtomicInteger(0);
            final AtomicInteger totalInvolvedAuthorities = new AtomicInteger(initialInvolvedAuthorities.size());
            final Set<Server> allInvolvedAuthorities = new HashSet<>(initialInvolvedAuthorities);
            final Set<Server> allActiveAuthorities = new HashSet<>();

            TaskExecutionFinishedCallback incrementFinishedCounterCallback = (task, taskExecution2, state) -> {
                int finishedCount = finishedTaskExecutionsCount.incrementAndGet();
                LOG.debug("Manifest #{} received of slice  '{}'", finishedCount, recoverSlice.getUrn());
            };

            for (Server authority : initialInvolvedAuthorities) {
                SliceManifestForAuthorityTask recoverSliverManifest =
                        hltf.getSliceManifest(recoverSlice, authority);

                TaskFinishedCallback checkActiveCallback = new TaskExecutionFinishedCallback() {
                    @Override
                    public void onTaskExecutionFinished(be.iminds.ilabt.jfed.highlevel.controller.Task task, TaskExecution taskExecution, TaskExecution.TaskState state) {
                        StringRspec stringRspec = recoverSliverManifest.getManifestStringRspec();
                        if (state == TaskExecution.TaskState.SUCCESS && stringRspec != null && ! stringRspec.getBasicNodeInfo().isEmpty()) {
                            allActiveAuthorities.add(authority);
                        }
                    }
                };

                recoverSliverManifest.addCallback(checkActiveCallback);
                taskThread.addTask(recoverSliverManifest, incrementFinishedCounterCallback);

                LOG.debug("Requesting manifest of slice  '{}' on '{}'", recoverSlice.getUrn(), authority.getName());
            }

            //wait for all tasks to finish
            while (finishedTaskExecutionsCount.get() < totalInvolvedAuthorities.get())
                try {
                    Thread.sleep(100);
                } catch (InterruptedException e) {
                    LOG.warn("Got interupted, so will abort recover");
                    return;
                }

            if (recoverSlice.getManifestRspec() != null) {
                recoverSlice.setRequestRspec(new RequestRspecSource(recoverSlice.getManifestRspec().getStringRspec(), ModelRspecType.BASIC));
                recoverSlice.getConnectAuthorities().addAll(allActiveAuthorities);
                LOG.debug("Recover success: Manifest is not null, and there are "+allActiveAuthorities.size()+" involved authorities");
            } else {
                LOG.info("Recovering slice failed: no manifest found on involved testbeds.");
                //will not see this as error. result will be Experiment in NON_EXISTING state
            }
        }


        experiment = new Experiment(recoverSlice, null);

        ExperimentControllerFactory experimentControllerFactory = injector.getInstance(ExperimentControllerFactory.class);
        experimentController = experimentControllerFactory.createExperimentController(experiment);

        experimentController.start();

        long maxWaitTimeMs = 9000l;
        long startTime = System.currentTimeMillis();
        while (experiment.getExperimentState() == ExperimentState.PENDING || experiment.getExperimentState() == ExperimentState.RESTORING) {
            try {
                Thread.sleep(100);
                long curTime = System.currentTimeMillis();
                long diff = curTime - startTime;
                if (diff >= maxWaitTimeMs) {
                    LOG.warn("Waiting for slice recovery took to long: aborting wait");
                    break;
                }
            } catch (InterruptedException e) {
                LOG.warn("Sleep interrupted. Will abort slice creation");
                return;
            }
        }

        if (experiment.getExperimentState() == ExperimentState.FAILED) {
            throw new WebApplicationException("Failed to create slice", Response.Status.INTERNAL_SERVER_ERROR);
        }

        assert experimentController != null;
        assert experiment != null;
        assert experiment.getSlice() != null;

        assert experimentController != null;
        assert experiment != null;
    }

    @Override
    public void renewSlice(Timestamp newExpire) {
        experimentController.renew(Instant.ofEpochMilli(newExpire.getTime()));
    }

    @Override
    public void deleteSlice() {
        experimentController.stop();

        long maxWaitTimeMs = 5000l;
        long startTime = System.currentTimeMillis();
        do {
            try {
                Thread.sleep(100);
                long curTime = System.currentTimeMillis();
                long diff = curTime - startTime;
                if (diff >= maxWaitTimeMs) {
                    LOG.warn("Waiting for delete took to long: aborting wait and returning current result.");
                    break;
                }
            } catch (InterruptedException e) {
                LOG.warn("Sleep interrupted. Will return current result");
                return;
            }
        } while (experiment.getExperimentState() == ExperimentState.DELETING);
    }

    @Override
    public String getProject() {
        if (experiment == null)
            return null;
        return experiment.getSubAuthName();
    }

    @Override
    public String getSliceName() {
        if (experiment == null)
            return null;
        return experiment.getName();
    }

    @Override
    public GeniUrn getSliceUrn() {
        if (experiment == null)
            return null;
        if (experiment.getSlice() == null)
            return null;
        return experiment.getSlice().getUrn();
    }

    @Override
    public Timestamp getExpire() {
        if (experiment == null)
            return null;
        if (experiment.getSlice() == null)
            return null;
        if (experiment.getSlice().getExpirationDate() == null)
            return null;
        return new Timestamp(experiment.getSlice().getExpirationDate().toEpochMilli());
    }

    @Override
    public String getManifestRspec() {
        if (experiment == null)
            return null;
        if (experiment.getSlice() == null)
            return null;
        ManifestRspecSource mrs = experiment.getSlice().getManifestRspec();
        if (mrs == null)
            return null;
        return mrs.getRspecXmlString();
    }


    @Override
    public Slice getExperimentSlice() {
        if (experiment == null)
            return null;
        return experiment.getSlice();
    }

    @Override
    public StatusDetails getExperimentSliceStatus() {
        if (experiment == null)
            return null;
        if (experiment.getSlice() == null)
            return null;
        return experiment.getSlice().getStatus();
    }

    @Override
    public Experiment getExperiment() {
        return experiment;
    }
    @Override
    public ExperimentState getExperimentState() {
        if (experiment == null)
            return null;
        return experiment.getExperimentState();
    }

    @Override
    public ExperimentController getExperimentController() {
        return experimentController;
    }


//    public String getLogXml() {
//        ByteArrayOutputStream outputStream = new ByteArrayOutputStream();
//
//        XMLOutputFactory xof = XMLOutputFactory.newInstance();
//        try {
//            XMLEventWriter xtw = new IndentingXMLEventWriter(xof.createXMLEventWriter(outputStream, "UTF-8"));
//            XMLEventFactory eventFactory = newInstance();
//
//            xtw.add(eventFactory.createStartDocument());
//
//            xtw.add(eventFactory.createStartElement(Q_BUGREPORT, null, null));
//
//            //write all the calls to the stream
//            xtw.add(eventFactory.createStartElement(Q_CALLS, null, null));
//
//            synchronized (javaFXTaskThread) {
//                for (JavaFXTaskThread.FXTaskExecution taskExecution : javaFXTaskThread.getAllTasks()) {
//                    for (SerializableApiCallDetails details : taskExecution.getApiCallHistory()) {
//
//                        SerializableApiCallDetailsWriter.writeApiCallDetails(details, xtw, eventFactory);
//                        xtw.flush();
//                    }
//                }
//            }
//
//            xtw.add(eventFactory.createEndElement(Q_CALLS, null)); //calls
//            xtw.add(eventFactory.createEndElement(Q_BUGREPORT, null)); //bugreport
//            xtw.flush();
//        } catch (XMLStreamException ex) {
//            throw new RuntimeException("Error while writing bug report to stream", ex);
//        }
//
//        String xmlLog = new String(outputStream.toByteArray(), Charset.forName("UTF-8"));
//        return xmlLog;
//    }
}
